"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.waitForHttpPortOpen = void 0;
const anylogger_1 = __importDefault(require("anylogger"));
const http_1 = __importDefault(require("http"));
const log = anylogger_1.default('nginx-testing:utils');
/**
 * Waits until the specified port is open and accepting an HTTP request, or the
 * `timeout` expires. Resolves to `true` on success, `false` on connection refused,
 * connection reset and connection timeout. Rejects on unexpected errors.
 *
 * @param requestOpts Request options; typically `hostname` and `port` should be provided.
 * @param timeoutMsec Connection timeout in milliseconds; attempts to open the socket will be
 *   terminated _after_ this time has passed.
 * @param intervalMsec Checking interval in milliseconds.
 */
const waitForHttpPortOpen = (requestOpts, timeoutMsec, intervalMsec = 100) => new Promise((resolve, reject) => {
    requestOpts = { method: 'HEAD', path: '/', ...requestOpts };
    const expectedErrorCodes = ['ECONNREFUSED', 'ECONNTIMEOUT', 'ECONNRESET'];
    const startTime = Date.now();
    const loop = () => {
        log.debug(`Trying to connect ${JSON.stringify(requestOpts)}`);
        return checkHttpWithTimeout(requestOpts, timeoutMsec)
            .then(() => resolve(true))
            .catch(err => {
            log.debug(`Got error ${err.code}: ${err}`);
            if (expectedErrorCodes.includes(err.code)) {
                if (Date.now() - startTime > timeoutMsec) {
                    log.debug(`Timed out after ${Date.now() - startTime}ms`);
                    return resolve(false);
                }
                else {
                    return setTimeout(loop, intervalMsec);
                }
            }
            return reject(err);
        });
    };
    loop();
});
exports.waitForHttpPortOpen = waitForHttpPortOpen;
const checkHttpWithTimeout = (opts, timeout) => new Promise((resolve, reject) => {
    const req = http_1.default.request({ ...opts, timeout });
    req
        .on('close', resolve)
        .on('error', (err) => {
        req.destroy();
        return reject(err);
    })
        .on('timeout', () => {
        req.abort();
        const err = Error('Connection timeout');
        err.code = 'ECONNTIMEOUT';
        return reject(err);
    })
        .end();
});
//# sourceMappingURL=waitForHttpPortOpen.js.map