/// <reference types="node" />
import type { Writable } from 'stream';
import { PatchOperation } from './nginxConf';
declare type ConfigPatch = Array<PatchOperation & {
    ifModule?: string;
}>;
/**
 * The default patch to be applied on the nginx configs to make it compatible with
 * the runner.
 */
export declare const configPatch: ConfigPatch;
/**
 * Options for the {@link startNginx} function.
 */
export declare type NginxOptions = RequiredKeys<BaseOptions, 'configPath'> | RequiredKeys<BaseOptions, 'config'>;
/**
 * Options for the {@link startNginx} function.
 */
interface BaseOptions {
    /**
     * Name or path of the nginx binary to start. Defaults to `'nginx'`.
     *
     * This option is ignored when `version` is provided.
     */
    binPath?: string;
    /**
     * A SemVer version range specifying the nginx version to run.
     *
     * Nginx binary for your OS and architecture will be downloaded from
     * [nginx-binaries](https://github.com/jirutka/nginx-binaries). It will be stored in
     * directory `.cache/nginx-binaries/` inside the nearest writable `node_modules`
     * directory or in `nginx-binaries/` inside the system-preferred temp directory.
     *
     * Not all versions are available. You can find a list of available binaries at
     * [nginx-binaries](https://jirutka.github.io/nginx-binaries/).
     */
    version?: string;
    /**
     * Nginx configuration to use.
     *
     * If `configPath` is provided, the processed config will be written to a temporary
     * file `.<filename>~` (where `<filename>` is a filename from `configPath`) in the
     * `configPath`'s directory (e.g. `conf/nginx.conf` -> `conf/.nginx.conf~`). Otherwise
     * it will be written into `nginx.conf` file in `workDir`. In either case, this file
     * will be automatically deleted after stopping the nginx.
     *
     * Either `configPath`, or `config` must be provided!
     */
    config?: string;
    /**
     * Path of the nginx configuration file to use.
     *
     * This file will be processed and the resulting config file will be written to
     * a temporary file `.<filename>~` (where `<filename>` is a filename from `configPath`)
     * in the `configPath`'s directory (e.g. `conf/nginx.conf` -> `conf/.nginx.conf~`).
     * This temporary file will be automatically deleted after stopping the nginx.
     *
     * Either `configPath`, or `config` must be provided!
     */
    configPath?: string;
    /**
     * Hostname or IP address to bind the port(s) on. Defaults to `'127.0.0.1'`.
     */
    bindAddress?: string;
    /**
     * A list of preferred port numbers to use for nginx.
     *
     * Unavailable ports (used by some other program or restricted by OS) are skipped.
     * If there are no preferred ports left and another port is needed, a random port
     * number is used.
     *
     * These ports are used to substitute `__PORT__`, `__PORT_1__`, ..., `__PORT_9__`
     * placeholders in the given nginx config.
     */
    preferredPorts?: number[];
    /**
     * Path of a directory that will be passed as a _prefix_ (`-p`) into `nginx`.
     * It will be automatically created if doesn't exist.
     *
     * If not provided, an unique temporary directory in the OS' default temp directory
     * (see `os.tmpdir()`) will be created and automatically deleted after stopping.
     */
    workDir?: string;
    /**
     * One of:
     *
     * - `'buffer'` -- Collect the nginx's stderr to a buffer that can be read using
     *   `readErrorLog()` (default).
     * - `'ignore'` - Ignore nginx's stderr.
     * - `'inherit'` -- Pass through the nginx's stderr output to the Node process.
     * - `<Writable>` -- A writable stream to pipe the nginx's stderr to.
     *
     * Nginx error log is expected to be redirected to _stderr_.
     * Directive `error_log stderr info;` will be automatically added to the config,
     * unless there's already `error_log` defined in the main context.
     */
    errorLog?: 'buffer' | 'ignore' | 'inherit' | Writable;
    /**
     * One of:
     *
     * - `'buffer'` -- Collect the nginx's access log to a buffer that can be read using
     *   `readAccessLog()` (default).
     * - `'ignore'` -- Ignore nginx's access log.
     * - `<Writable>` -- A writable stream to pipe the nginx's access log to.
     *
     * Nginx access log is expected to be redirected to file `<workDir>/access.log`.
     * Directive `access_log access.log;` will be automatically added to the config,
     * unless there's already `access_log` defined in the `http` context.
     */
    accessLog?: 'buffer' | 'ignore' | Writable;
    /**
     * Number of milliseconds after the start to wait for the nginx to respond to the
     * health-check request (`HEAD http://<bindAddress>:<ports[0]>/`).
     *
     * Defaults to `1000`.
     */
    startTimeoutMsec?: number;
    /**
     * The logger to use for the runner logging. It must be an object with functions:
     * `debug`, `info`, `warn`, and `error`. You can use even global `console` as the
     * logger.
     *
     * Defaults to [anylogger](https://github.com/download/anylogger) with logger name
     * `'nginx-testing'`.
     */
    logger?: Logger;
}
declare type Logger = Pick<Console, 'debug' | 'info' | 'warn' | 'error'>;
declare type RequiredKeys<O extends object, K extends keyof O = keyof O> = O & {
    [L in K]-?: O[L];
};
/**
 * A return value of the {@link startNginx} function.
 */
export interface NginxServer {
    /**
     * The current nginx configuration.
     */
    readonly config: string;
    /**
     * PID of the nginx process.
     */
    readonly pid: number;
    /**
     * Number of the first port allocated for nginx, i.e. the port on which nginx
     * should listen for connections. It's the same as `ports[0]`.
     */
    readonly port: number;
    /**
     * A list of port numbers allocated for nginx.
     */
    readonly ports: ReadonlyArray<number>;
    /**
     * Path of the nginx's working directory.
     */
    readonly workDir: string;
    /**
     * Reads new messages from the access log since the last call of `readAccessLog()`.
     *
     * @throws {Error} if the process was created with option `accessLog` other than
     *   `'buffer'` or `undefined`.
     */
    readAccessLog(): Promise<string>;
    /**
     * Reads new messages from the error log since the last call of `readErrorLog()`.
     *
     * @throws {Error} if the process was created with option `errorLog` other than
     *   `'buffer'` or `undefined`.
     */
    readErrorLog(): Promise<string>;
    /**
     * Stops the nginx and cleans-up temporary files and directories.
     */
    stop(): Promise<void>;
}
/**
 * Starts nginx server with the given configuration.
 *
 * @example
 * import { startNginx, NginxProcess } from 'nginx-testing'
 * import fetch from 'node-fetch'
 *
 * let nginx: NginxProcess
 *
 * before(async () => {
 *   nginx = await startNginx({ version: '1.18.x', configPath: './nginx.conf' })
 * })
 * after(nginx.stop)
 *
 * test('GET / results in HTTP 200', async () => {
 *   const resp = await fetch(`http://localhost:${nginx.port}/`)
 *   assert(resp.status === 200)
 * })
 */
export declare function startNginx(opts: NginxOptions): Promise<NginxServer>;
export {};
//# sourceMappingURL=nginxRunner.d.ts.map