"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.__testing = exports.startNginx = exports.configPatch = void 0;
const anylogger_1 = __importDefault(require("anylogger"));
const path_1 = __importDefault(require("path"));
const tail_file_1 = __importDefault(require("@logdna/tail-file"));
const execa_1 = __importDefault(require("execa"));
const get_port_1 = __importDefault(require("get-port"));
const nginx_binaries_1 = require("nginx-binaries");
const stream_buffers_1 = require("stream-buffers");
const FS = __importStar(require("./internal/fs"));
const useCleanup_1 = require("./internal/useCleanup");
const tempDir_1 = require("./internal/tempDir");
const waitForHttpPortOpen_1 = require("./internal/waitForHttpPortOpen");
const nginxConf_1 = require("./nginxConf");
const nginxVersionInfo_1 = require("./nginxVersionInfo");
const defaultLogger = anylogger_1.default('nginx-testing');
/**
 * The default patch to be applied on the nginx configs to make it compatible with
 * the runner.
 */
exports.configPatch = [
    { path: '/daemon', op: 'set', value: 'off' },
    { path: '/pid', op: 'set', value: 'nginx.pid' },
    // This is necessary on Windows, otherwise execa fails to kill nginx.
    { path: '/master_process', op: 'default', value: 'off' },
    { path: '/error_log', op: 'default', value: 'stderr info' },
    { path: '/http/access_log', op: 'default', value: 'access.log' },
    { path: '/http/client_body_temp_path', op: 'default', value: 'client_body_temp' },
    { path: '/http/proxy_temp_path', op: 'default', value: 'proxy_temp', ifModule: 'http_proxy' },
    { path: '/http/fastcgi_temp_path', op: 'default', value: 'fastcgi_temp', ifModule: 'http_fastcgi' },
    { path: '/http/uwsgi_temp_path', op: 'default', value: 'uwsgi_temp', ifModule: 'http_uwsgi' },
    { path: '/http/scgi_temp_path', op: 'default', value: 'scgi_temp', ifModule: 'http_scgi' },
];
/**
 * Starts nginx server with the given configuration.
 *
 * @example
 * import { startNginx, NginxProcess } from 'nginx-testing'
 * import fetch from 'node-fetch'
 *
 * let nginx: NginxProcess
 *
 * before(async () => {
 *   nginx = await startNginx({ version: '1.18.x', configPath: './nginx.conf' })
 * })
 * after(nginx.stop)
 *
 * test('GET / results in HTTP 200', async () => {
 *   const resp = await fetch(`http://localhost:${nginx.port}/`)
 *   assert(resp.status === 200)
 * })
 */
async function startNginx(opts) {
    var _a;
    if (!opts.config && !opts.configPath) {
        throw TypeError('Either config or configPath must be provided');
    }
    const { accessLog = 'buffer', bindAddress = '127.0.0.1', errorLog = 'buffer', logger: log = defaultLogger, preferredPorts = [], startTimeoutMsec = 1000, } = opts;
    const [onCleanup, cleanup] = useCleanup_1.useCleanup({ registerExitHook: true });
    try {
        let workDir = opts.workDir;
        if (workDir) {
            await FS.mkdir(workDir, { recursive: true });
        }
        else {
            workDir = tempDir_1.createTempDir('nginx-testing');
            // Async rm does not remove the dir on Windows (and I have no idea why).
            onCleanup(() => FS.rmRfSync(workDir));
        }
        const binPath = opts.version
            ? await nginx_binaries_1.NginxBinary.download({ version: opts.version })
            : (opts.binPath || 'nginx');
        // Prepare config
        let config = (_a = opts.config) !== null && _a !== void 0 ? _a : await FS.readFile(opts.configPath, 'utf8');
        const portsCount = countNeededPorts(config);
        if (portsCount < 1 && preferredPorts.length < 1) {
            throw Error('No __PORT__ placeholder found in nginx config and option preferredPorts is empty');
        }
        const versionInfo = await nginxVersionInfo_1.nginxVersionInfo(binPath);
        const ports = await getFreePorts(bindAddress, countNeededPorts(config) || 1, preferredPorts);
        config = adjustConfig(config, { ...versionInfo, ports, workDir });
        const configPath = opts.configPath
            ? tempConfigPath(opts.configPath)
            : path_1.default.join(workDir, 'nginx.conf');
        log.debug(`Writing config to ${configPath}:\n-----BEGIN CONFIG-----\n${config}\n-----END CONFIG-----`);
        await FS.writeFile(configPath, config, 'utf8');
        onCleanup(() => FS.rmRfSync(configPath));
        // Start nginx
        log.info(`Starting nginx ${versionInfo.version} on port(s): ${ports.join(', ')}`);
        const ngxProcess = execa_1.default(binPath, ['-c', configPath, '-p', workDir], {
            stdin: 'ignore',
            stdout: 'ignore',
            stderr: errorLog === 'buffer' ? 'pipe' : errorLog,
        });
        onCleanup(() => {
            log.debug(`Stopping nginx (${ngxProcess.pid})`);
            ngxProcess.cancel();
        });
        log.debug(`Nginx started with PID ${ngxProcess.pid}`);
        // Set-up error log
        let errorLogBuffer;
        if (errorLog === 'buffer') {
            errorLogBuffer = new stream_buffers_1.WritableStreamBuffer();
            ngxProcess.stderr.pipe(errorLogBuffer);
        }
        // Check if running
        // If nginx cannot be executed, e.g. invalid path, we want to fail fast
        // and throw a relevant error.
        await new Promise((resolve, reject) => {
            ngxProcess.once('error', reject);
            setTimeout(resolve, 50); // sleep up to 50 ms
        });
        if (!await waitForHttpPortOpen_1.waitForHttpPortOpen({ hostname: bindAddress, port: ports[0] }, startTimeoutMsec)) {
            const msg = errorLogBuffer === null || errorLogBuffer === void 0 ? void 0 : errorLogBuffer.getContentsAsString();
            msg && log.error(msg);
            throw Error(`Failed to start nginx, no response on port ${ports[0]}`);
        }
        // Set-up access log
        let accessLogTail;
        let accessLogBuffer;
        if (accessLog !== 'ignore') {
            accessLogTail = new tail_file_1.default(path_1.default.join(workDir, 'access.log'), { pollFileIntervalMs: 10 });
            accessLogTail.pipe(accessLog === 'buffer'
                ? (accessLogBuffer = new stream_buffers_1.WritableStreamBuffer())
                : accessLog);
            log.debug(`Begins polling of ${path_1.default.join(workDir, 'access.log')}`);
            await accessLogTail.start();
            onCleanup(async () => await accessLogTail.quit());
        }
        // Return
        return {
            get config() { return config; },
            pid: ngxProcess.pid,
            ports,
            port: ports[0],
            workDir,
            readAccessLog: async () => {
                if (!accessLogBuffer || !accessLogTail) {
                    throw Error("This function is available only when the option 'accessLog' is 'buffer'");
                }
                if ('_pollFileForChanges' in accessLogTail) {
                    await accessLogTail._pollFileForChanges();
                }
                return accessLogBuffer.getContentsAsString() || '';
            },
            // This function doesn't need to be async now, but may be in the future.
            readErrorLog: async () => {
                if (!errorLogBuffer) {
                    throw Error("This function is available only when the option 'errorLog' is 'buffer'");
                }
                return errorLogBuffer.getContentsAsString() || '';
            },
            stop: cleanup,
        };
    }
    catch (err) {
        await cleanup();
        throw err;
    }
}
exports.startNginx = startNginx;
const portPlaceholderRx = /\b__PORT(?:_(\d))?__\b/g;
function countNeededPorts(config) {
    const portIndexes = Array.from(config.matchAll(portPlaceholderRx), ([_, n]) => Number(n || 0));
    return Math.max(...portIndexes) + 1;
}
async function getFreePorts(address, count, preferred = []) {
    const ports = [];
    for (let i = count; i > 0; i--) {
        ports.push(await get_port_1.default({ host: address, port: preferred }));
    }
    return ports;
}
function adjustConfig(config, { modules, ports, workDir }) {
    config = config
        // nginx requires forward slashes even on Windows.
        .replace(/\b__WORKDIR__\b/g, workDir.replace(/\\/g, '/'))
        .replace(portPlaceholderRx, (match, idx) => { var _a, _b; return (_b = (_a = ports[Number(idx) || 0]) === null || _a === void 0 ? void 0 : _a.toString()) !== null && _b !== void 0 ? _b : match; });
    const patch = exports.configPatch.filter(({ ifModule }) => {
        return !ifModule || modules[ifModule] !== 'without' && modules[ifModule] !== 'with-dynamic';
    });
    if (patch.length > 0) {
        config = nginxConf_1.parseConf(config).applyPatch(patch).toString();
    }
    return config;
}
function tempConfigPath(filepath) {
    return path_1.default.join(path_1.default.dirname(path_1.default.resolve(filepath)), `.${path_1.default.basename(filepath)}~`);
}
/** @internal */
exports.__testing = {
    adjustConfig,
};
//# sourceMappingURL=nginxRunner.js.map